@@ -1,4 +1,3 @@
-
 #include <qpdf/QPDF.hh>
 
 #include <vector>
@@ -16,6 +15,8 @@
 #include <qpdf/QPDF_Null.hh>
 #include <qpdf/QPDF_Dictionary.hh>
 
+std::string QPDF::qpdf_version = "2.1.2";
+
 void
 QPDF::InputSource::setLastOffset(off_t offset)
 {
@@ -141,7 +142,9 @@ QPDF::FileInputSource::read(char* buffer, int length)
     size_t len = fread(buffer, 1, length, this->file);
     if ((len == 0) && ferror(this->file))
     {
-	throw QPDFExc(this->filename, this->last_offset,
+	throw QPDFExc(qpdf_e_system,
+		      this->filename, "",
+		      this->last_offset,
 		      std::string("read ") +
 		      QUtil::int_to_string(length) + " bytes");
     }
@@ -197,7 +200,8 @@ QPDF::BufferInputSource::seek(off_t offset, int whence)
 	break;
 
       default:
-	throw QEXC::Internal("invalid argument to BufferInputSource::seek");
+	throw std::logic_error(
+	    "INTERNAL ERROR: invalid argument to BufferInputSource::seek");
 	break;
     }
 }
@@ -241,18 +245,29 @@ QPDF::ObjGen::ObjGen(int o = 0, int g = 0) :
 }
 
 bool
-QPDF::ObjGen::ObjGen::operator<(ObjGen const& rhs) const
+QPDF::ObjGen::operator<(ObjGen const& rhs) const
 {
     return ((this->obj < rhs.obj) ||
 	    ((this->obj == rhs.obj) && (this->gen < rhs.gen)));
 }
 
+std::string const&
+QPDF::QPDFVersion()
+{
+    return QPDF::qpdf_version;
+}
+
 QPDF::QPDF() :
     encrypted(false),
     encryption_initialized(false),
     ignore_xref_streams(false),
     suppress_warnings(false),
     attempt_recovery(true),
+    encryption_V(0),
+    encrypt_metadata(true),
+    cf_stream(e_none),
+    cf_string(e_none),
+    cf_file(e_none),
     cached_key_objid(0),
     cached_key_generation(0),
     first_xref_item_offset(0),
@@ -268,7 +283,10 @@ void
 QPDF::processFile(char const* filename, char const* password)
 {
     this->file.setFilename(filename);
-    this->provided_password = password;
+    if (password)
+    {
+	this->provided_password = password;
+    }
     parse();
 }
 
@@ -290,10 +308,10 @@ QPDF::setAttemptRecovery(bool val)
     this->attempt_recovery = val;
 }
 
-std::vector<std::string>
+std::vector<QPDFExc>
 QPDF::getWarnings()
 {
-    std::vector<std::string> result = this->warnings;
+    std::vector<QPDFExc> result = this->warnings;
     this->warnings.clear();
     return result;
 }
@@ -317,7 +335,8 @@ QPDF::parse()
     else
     {
 	QTC::TC("qpdf", "QPDF not a pdf file");
-	throw QPDFExc(this->file.getName(), 0, "not a PDF file");
+	throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+		      "", 0, "not a PDF file");
     }
 
     // PDF spec says %%EOF must be found within the last 1024 bytes of
@@ -361,7 +380,8 @@ QPDF::parse()
 	if (! m2)
 	{
 	    QTC::TC("qpdf", "QPDF can't find startxref");
-	    throw QPDFExc(this->file.getName() + ": can't find startxref");
+	    throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(), "", 0,
+			  "can't find startxref");
 	}
 	off_t xref_offset = atoi(m2.getMatch(1).c_str());
 	read_xref(xref_offset);
@@ -385,10 +405,10 @@ QPDF::parse()
 void
 QPDF::warn(QPDFExc const& e)
 {
-    this->warnings.push_back(e.unparse());
+    this->warnings.push_back(e);
     if (! this->suppress_warnings)
     {
-	std::cerr << "WARNING: " << this->warnings.back() << std::endl;
+	std::cerr << "WARNING: " << this->warnings.back().what() << std::endl;
     }
 }
 
@@ -405,13 +425,32 @@ QPDF::setTrailer(QPDFObjectHandle obj)
 void
 QPDF::reconstruct_xref(QPDFExc& e)
 {
-    static PCRE obj_re("^(\\d+) (\\d+) obj\\b");
-    static PCRE endobj_re("^endobj\\b");
-    static PCRE trailer_re("^trailer\\b");
+    static PCRE obj_re("^\\s*(\\d+)\\s+(\\d+)\\s+obj\\b");
+    static PCRE endobj_re("^\\s*endobj\\b");
+    static PCRE trailer_re("^\\s*trailer\\b");
 
-    warn(QPDFExc(this->file.getName(), 0, "file is damaged"));
+    warn(QPDFExc(qpdf_e_damaged_pdf, this->file.getName(), "", 0,
+		 "file is damaged"));
     warn(e);
-    warn(QPDFExc("Attempting to reconstruct cross-reference table"));
+    warn(QPDFExc(qpdf_e_damaged_pdf, this->file.getName(), "", 0,
+		 "Attempting to reconstruct cross-reference table"));
+
+    // Delete all references to type 1 (uncompressed) objects
+    std::set<ObjGen> to_delete;
+    for (std::map<ObjGen, QPDFXRefEntry>::iterator iter =
+	     this->xref_table.begin();
+	 iter != this->xref_table.end(); ++iter)
+    {
+	if (((*iter).second).getType() == 1)
+	{
+	    to_delete.insert((*iter).first);
+	}
+    }
+    for (std::set<ObjGen>::iterator iter = to_delete.begin();
+	 iter != to_delete.end(); ++iter)
+    {
+	this->xref_table.erase(*iter);
+    }
 
     this->file.seek(0, SEEK_END);
     off_t eof = this->file.tell();
@@ -444,7 +483,8 @@ QPDF::reconstruct_xref(QPDFExc& e)
 		// read "trailer"
 		this->file.seek(this->file.getLastOffset(), SEEK_SET);
 		readToken(&this->file);
-		QPDFObjectHandle t = readObject(&this->file, 0, 0, false);
+		QPDFObjectHandle t =
+		    readObject(&this->file, "trailer", 0, 0, false);
 		if (! t.isDictionary())
 		{
 		    // Oh well.  It was worth a try.
@@ -465,7 +505,8 @@ QPDF::reconstruct_xref(QPDFExc& e)
 	// with bad startxref pointers even when they have object
 	// streams.
 
-	throw QPDFExc(this->file.getName() + ": unable to find trailer "
+	throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(), "", 0,
+		      "unable to find trailer "
 		      "dictionary while recovering damaged file");
     }
 
@@ -505,8 +546,8 @@ QPDF::read_xref(off_t xref_offset)
     if (size != max_obj + 1)
     {
 	QTC::TC("qpdf", "QPDF xref size mismatch");
-	warn(QPDFExc(this->file.getName() +
-		     std::string(": reported number of objects (") +
+	warn(QPDFExc(qpdf_e_damaged_pdf, this->file.getName(), "", 0,
+		     std::string("reported number of objects (") +
 		     QUtil::int_to_string(size) +
 		     ") inconsistent with actual number of objects (" +
 		     QUtil::int_to_string(max_obj + 1) + ")"));
@@ -520,7 +561,7 @@ QPDF::read_xref(off_t xref_offset)
 int
 QPDF::read_xrefTable(off_t xref_offset)
 {
-    static PCRE xref_first_re("^(\\d+)\\s+(\\d+)");
+    static PCRE xref_first_re("^\\s*(\\d+)\\s+(\\d+)");
     static PCRE xref_entry_re("(?s:(^\\d{10}) (\\d{5}) ([fn])[ \r\n]{2}$)");
 
     std::vector<ObjGen> deleted_items;
@@ -534,7 +575,8 @@ QPDF::read_xrefTable(off_t xref_offset)
 	if (! m1)
 	{
 	    QTC::TC("qpdf", "QPDF invalid xref");
-	    throw QPDFExc(this->file.getName(), this->file.getLastOffset(),
+	    throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+			  "xref table", this->file.getLastOffset(),
 			  "xref syntax invalid");
 	}
 	int obj = atoi(m1.getMatch(1).c_str());
@@ -555,7 +597,8 @@ QPDF::read_xrefTable(off_t xref_offset)
 	    {
 		QTC::TC("qpdf", "QPDF invalid xref entry");
 		throw QPDFExc(
-		    this->file.getName(), this->file.getLastOffset(),
+		    qpdf_e_damaged_pdf, this->file.getName(),
+		    "xref table", this->file.getLastOffset(),
 		    "invalid xref entry (obj=" +
 		    QUtil::int_to_string(i) + ")");
 	    }
@@ -587,11 +630,13 @@ QPDF::read_xrefTable(off_t xref_offset)
     }
 
     // Set offset to previous xref table if any
-    QPDFObjectHandle cur_trailer = readObject(&this->file, 0, 0, false);
+    QPDFObjectHandle cur_trailer =
+	readObject(&this->file, "trailer", 0, 0, false);
     if (! cur_trailer.isDictionary())
     {
 	QTC::TC("qpdf", "QPDF missing trailer");
-	throw QPDFExc(this->file.getName(), this->file.getLastOffset(),
+	throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+		      "", this->file.getLastOffset(),
 		      "expected trailer dictionary");
     }
 
@@ -602,13 +647,15 @@ QPDF::read_xrefTable(off_t xref_offset)
 	if (! this->trailer.hasKey("/Size"))
 	{
 	    QTC::TC("qpdf", "QPDF trailer lacks size");
-	    throw QPDFExc(this->file.getName(), this->file.getLastOffset(),
+	    throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+			  "trailer", this->file.getLastOffset(),
 			  "trailer dictionary lacks /Size key");
 	}
 	if (! this->trailer.getKey("/Size").isInteger())
 	{
 	    QTC::TC("qpdf", "QPDF trailer size not integer");
-	    throw QPDFExc(this->file.getName(), this->file.getLastOffset(),
+	    throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+			  "trailer", this->file.getLastOffset(),
 			  "/Size key in trailer dictionary is not "
 			  "an integer");
 	}
@@ -632,7 +679,8 @@ QPDF::read_xrefTable(off_t xref_offset)
 	    }
 	    else
 	    {
-		throw QPDFExc(this->file.getName(), xref_offset,
+		throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+			      "xref stream", xref_offset,
 			      "invalid /XRefStm");
 	    }
 	}
@@ -651,7 +699,8 @@ QPDF::read_xrefTable(off_t xref_offset)
 	if (! cur_trailer.getKey("/Prev").isInteger())
 	{
 	    QTC::TC("qpdf", "QPDF trailer prev not integer");
-	    throw QPDFExc(this->file.getName(), this->file.getLastOffset(),
+	    throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+			  "trailer", this->file.getLastOffset(),
 			  "/Prev key in trailer dictionary is not "
 			  "an integer");
 	}
@@ -677,7 +726,8 @@ QPDF::read_xrefStream(off_t xref_offset)
 	QPDFObjectHandle xref_obj;
 	try
 	{
-	    xref_obj = readObjectAtOffset(xref_offset, 0, 0, xobj, xgen);
+	    xref_obj = readObjectAtOffset(
+		false, xref_offset, "xref stream", 0, 0, xobj, xgen);
 	}
 	catch (QPDFExc& e)
 	{
@@ -697,7 +747,8 @@ QPDF::read_xrefStream(off_t xref_offset)
     if (! found)
     {
 	QTC::TC("qpdf", "QPDF can't find xref");
-	throw QPDFExc(this->file.getName(), xref_offset, "xref not found");
+	throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+		      "", xref_offset, "xref not found");
     }
 
     return xref_offset;
@@ -717,7 +768,8 @@ QPDF::processXRefStream(off_t xref_offset, QPDFObjectHandle& xref_obj)
 	   dict.getKey("/Size").isInteger() &&
 	   (Index_obj.isArray() || Index_obj.isNull())))
     {
-	throw QPDFExc(this->file.getName(), xref_offset,
+	throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+		      "xref stream", xref_offset,
 		      "Cross-reference stream does not have"
 		      " proper /W and /Index keys");
     }
@@ -727,7 +779,8 @@ QPDF::processXRefStream(off_t xref_offset, QPDFObjectHandle& xref_obj)
 	int n_index = Index_obj.getArrayNItems();
 	if ((n_index % 2) || (n_index < 2))
 	{
-	    throw QPDFExc(this->file.getName(), xref_offset,
+	    throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+			  "xref stream", xref_offset,
 			  "Cross-reference stream's /Index has an"
 			  " invalid number of values");
 	}
@@ -739,7 +792,8 @@ QPDF::processXRefStream(off_t xref_offset, QPDFObjectHandle& xref_obj)
 	    }
 	    else
 	    {
-		throw QPDFExc(this->file.getName(), xref_offset,
+		throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+			      "xref stream", xref_offset,
 			      "Cross-reference stream's /Index's item " +
 			      QUtil::int_to_string(i) +
 			      " is not an integer");
@@ -777,10 +831,19 @@ QPDF::processXRefStream(off_t xref_offset, QPDFObjectHandle& xref_obj)
 
     if (expected_size != actual_size)
     {
-	throw QPDFExc(this->file.getName(), xref_offset,
-		      "Cross-reference stream data has the wrong size;"
-		      " expected = " + QUtil::int_to_string(expected_size) +
-		      "; actual = " + QUtil::int_to_string(actual_size));
+	QPDFExc x(qpdf_e_damaged_pdf, this->file.getName(),
+		  "xref stream", xref_offset,
+		  "Cross-reference stream data has the wrong size;"
+		  " expected = " + QUtil::int_to_string(expected_size) +
+		  "; actual = " + QUtil::int_to_string(actual_size));
+	if (expected_size > actual_size)
+	{
+	    throw x;
+	}
+	else
+	{
+	    warn(x);
+	}
     }
 
     int cur_chunk = 0;
@@ -850,7 +913,8 @@ QPDF::processXRefStream(off_t xref_offset, QPDFObjectHandle& xref_obj)
     {
 	if (! dict.getKey("/Prev").isInteger())
 	{
-	    throw QPDFExc(this->file.getName(), this->file.getLastOffset(),
+	    throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+			  "xref stream", this->file.getLastOffset(),
 			  "/Prev key in xref stream dictionary is not "
 			  "an integer");
 	}
@@ -919,7 +983,8 @@ QPDF::insertXrefEntry(int obj, int f0, int f1, int f2, bool overwrite)
 	break;
 
       default:
-	throw QPDFExc(this->file.getName(), 0,
+	throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+		      "xref stream", this->file.getLastOffset(),
 		      "unknown xref stream entry type " +
 		      QUtil::int_to_string(f0));
 	break;
@@ -948,18 +1013,40 @@ QPDF::showXRefTable()
 	    break;
 
 	  default:
-	    throw QEXC::Internal("unknown cross-reference table type while"
-				 " showing xref_table");
+	    throw std::logic_error("unknown cross-reference table type while"
+				   " showing xref_table");
 	    break;
 	}
 	std::cout << std::endl;
     }
 }
 
-QPDFObjectHandle
-QPDF::readObject(InputSource* input, int objid, int generation,
-		 bool in_object_stream)
+void
+QPDF::setLastObjectDescription(std::string const& description,
+			       int objid, int generation)
 {
+    this->last_object_description.clear();
+    if (! description.empty())
+    {
+	this->last_object_description += description;
+	if (objid > 0)
+	{
+	    this->last_object_description += ": ";
+	}
+    }
+    if (objid > 0)
+    {
+	this->last_object_description += "object " +
+	    QUtil::int_to_string(objid) + " " +
+	    QUtil::int_to_string(generation);
+    }
+}
+
+QPDFObjectHandle
+QPDF::readObject(InputSource* input, std::string const& description,
+		 int objid, int generation, bool in_object_stream)
+{
+    setLastObjectDescription(description, objid, generation);
     off_t offset = input->tell();
     QPDFObjectHandle object = readObjectInternal(
 	input, objid, generation, in_object_stream, false, false);
@@ -980,7 +1067,8 @@ QPDF::readObjectInternal(InputSource* input,
 	// Although dictionaries and arrays arbitrarily nest, these
 	// variables indicate what is at the top of the stack right
 	// now, so they can, by definition, never both be true.
-	throw QEXC::Internal("readObjectInternal: in_dict && in_array");
+	throw std::logic_error(
+	    "INTERNAL ERROR: readObjectInternal: in_dict && in_array");
     }
 
     QPDFObjectHandle object;
@@ -1000,7 +1088,9 @@ QPDF::readObjectInternal(InputSource* input,
 	  case QPDFTokenizer::tt_brace_close:
 	    // Don't know what to do with these for now
 	    QTC::TC("qpdf", "QPDF bad brace");
-	    throw QPDFExc(input->getName(), input->getLastOffset(),
+	    throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+			  this->last_object_description,
+			  input->getLastOffset(),
 			  "unexpected brace token");
 	    break;
 
@@ -1012,7 +1102,9 @@ QPDF::readObjectInternal(InputSource* input,
 	    else
 	    {
 		QTC::TC("qpdf", "QPDF bad array close");
-		throw QPDFExc(input->getName(), input->getLastOffset(),
+		throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+			      this->last_object_description,
+			      input->getLastOffset(),
 			      "unexpected array close token");
 	    }
 	    break;
@@ -1025,7 +1117,9 @@ QPDF::readObjectInternal(InputSource* input,
 	    else
 	    {
 		QTC::TC("qpdf", "QPDF bad dictionary close");
-		throw QPDFExc(input->getName(), input->getLastOffset(),
+		throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+			      this->last_object_description,
+			      input->getLastOffset(),
 			      "unexpected dictionary close token");
 	    }
 	    break;
@@ -1080,7 +1174,9 @@ QPDF::readObjectInternal(InputSource* input,
 		}
 		else
 		{
-		    throw QPDFExc(input->getName(), input->getLastOffset(),
+		    throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+				  this->last_object_description,
+				  input->getLastOffset(),
 				  "unknown token while reading object (" +
 				  value + ")");
 		}
@@ -1099,7 +1195,9 @@ QPDF::readObjectInternal(InputSource* input,
 	    break;
 
 	  default:
-	    throw QPDFExc(input->getName(), input->getLastOffset(),
+	    throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+			  this->last_object_description,
+			  input->getLastOffset(),
 			  "unknown token type while reading object");
 	    break;
 	}
@@ -1113,9 +1211,10 @@ QPDF::readObjectInternal(InputSource* input,
 	}
 	else if (! object.isInitialized())
 	{
-	    throw QEXC::Internal(std::string("uninitialized object (token = ") +
-				 QUtil::int_to_string(token.getType()) +
-				 ", " + token.getValue() + ")");
+	    throw std::logic_error(
+		"INTERNAL ERROR: uninitialized object (token = " +
+		QUtil::int_to_string(token.getType()) +
+		", " + token.getValue() + ")");
 	}
 	else
 	{
@@ -1135,7 +1234,8 @@ QPDF::readObjectInternal(InputSource* input,
 	{
 	    QTC::TC("qpdf", "QPDF dictionary odd number of elements");
 	    throw QPDFExc(
-		input->getName(), input->getLastOffset(),
+		qpdf_e_damaged_pdf, input->getName(),
+		this->last_object_description, input->getLastOffset(),
 		"dictionary ending here has an odd number of elements");
 	}
 	for (unsigned int i = 0; i < olist.size(); i += 2)
@@ -1145,7 +1245,8 @@ QPDF::readObjectInternal(InputSource* input,
 	    if (! key_obj.isName())
 	    {
 		throw QPDFExc(
-		    input->getName(), offset,
+		    qpdf_e_damaged_pdf,
+		    input->getName(), this->last_object_description, offset,
 		    std::string("dictionary key not name (") +
 		    key_obj.unparse() + ")");
 	    }
@@ -1191,7 +1292,8 @@ QPDF::readObjectInternal(InputSource* input,
 		    if (dict.count("/Length") == 0)
 		    {
 			QTC::TC("qpdf", "QPDF stream without length");
-			throw QPDFExc(input->getName(), offset,
+			throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+				      this->last_object_description, offset,
 				      "stream dictionary lacks /Length key");
 		    }
 
@@ -1199,7 +1301,8 @@ QPDF::readObjectInternal(InputSource* input,
 		    if (! length_obj.isInteger())
 		    {
 			QTC::TC("qpdf", "QPDF stream length not integer");
-			throw QPDFExc(input->getName(), offset,
+			throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+				      this->last_object_description, offset,
 				      "/Length key in stream dictionary is not "
 				      "an integer");
 		    }
@@ -1211,7 +1314,9 @@ QPDF::readObjectInternal(InputSource* input,
 			       QPDFTokenizer::tt_word, "endstream")))
 		    {
 			QTC::TC("qpdf", "QPDF missing endstream");
-			throw QPDFExc(input->getName(), input->getLastOffset(),
+			throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+				      this->last_object_description,
+				      input->getLastOffset(),
 				      "expected endstream");
 		    }
 		}
@@ -1245,11 +1350,12 @@ int
 QPDF::recoverStreamLength(InputSource* input,
 			  int objid, int generation, off_t stream_offset)
 {
-    static PCRE endobj_re("^endobj\\b");
+    static PCRE endobj_re("^\\s*endobj\\b");
 
     // Try to reconstruct stream length by looking for
     // endstream(\r\n?|\n)endobj
-    warn(QPDFExc(input->getName(), stream_offset,
+    warn(QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+		 this->last_object_description, stream_offset,
 		 "attempting to recover stream length"));
 
     input->seek(0, SEEK_END);
@@ -1318,7 +1424,8 @@ QPDF::recoverStreamLength(InputSource* input,
 
     if (length == 0)
     {
-	throw QPDFExc(input->getName(), stream_offset,
+	throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+		      this->last_object_description, stream_offset,
 		      "unable to recover stream data");
     }
 
@@ -1338,11 +1445,14 @@ QPDF::readToken(InputSource* input)
 	char ch;
 	if (input->read(&ch, 1) == 0)
 	{
-	    throw QPDFExc(input->getName(), offset, "EOF while reading token");
+	    throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+			  this->last_object_description, offset,
+			  "EOF while reading token");
 	}
 	else
 	{
-	    if (isspace(ch) && (input->getLastOffset() == offset))
+	    if (isspace((unsigned char)ch) &&
+		(input->getLastOffset() == offset))
 	    {
 		++offset;
 	    }
@@ -1357,7 +1467,9 @@ QPDF::readToken(InputSource* input)
 
     if (token.getType() == QPDFTokenizer::tt_bad)
     {
-	throw QPDFExc(input->getName(), offset, token.getErrorMessage());
+	throw QPDFExc(qpdf_e_damaged_pdf, input->getName(),
+		      this->last_object_description, offset,
+		      token.getErrorMessage());
     }
 
     input->setLastOffset(offset);
@@ -1366,9 +1478,12 @@ QPDF::readToken(InputSource* input)
 }
 
 QPDFObjectHandle
-QPDF::readObjectAtOffset(off_t offset, int exp_objid, int exp_generation,
+QPDF::readObjectAtOffset(bool try_recovery,
+			 off_t offset, std::string const& description,
+			 int exp_objid, int exp_generation,
 			 int& objid, int& generation)
 {
+    setLastObjectDescription(description, exp_objid, exp_generation);
     this->file.seek(offset, SEEK_SET);
 
     QPDFTokenizer::Token tobjid = readToken(&this->file);
@@ -1388,7 +1503,9 @@ QPDF::readObjectAtOffset(off_t offset, int exp_objid, int exp_generation,
 	if (! (objidok && genok && objok))
 	{
 	    QTC::TC("qpdf", "QPDF expected n n obj");
-	    throw QPDFExc(this->file.getName(), offset, "expected n n obj");
+	    throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+			  this->last_object_description, offset,
+			  "expected n n obj");
 	}
 	objid = atoi(tobjid.getValue().c_str());
 	generation = atoi(tgen.getValue().c_str());
@@ -1397,7 +1514,8 @@ QPDF::readObjectAtOffset(off_t offset, int exp_objid, int exp_generation,
 	    (! ((objid == exp_objid) && (generation == exp_generation))))
 	{
 	    QTC::TC("qpdf", "QPDF err wrong objid/generation");
-	    throw QPDFExc(this->file.getName(), offset,
+	    throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+			  this->last_object_description, offset,
 			  std::string("expected ") +
 			  QUtil::int_to_string(exp_objid) + " " +
 			  QUtil::int_to_string(exp_generation) + " obj");
@@ -1405,7 +1523,7 @@ QPDF::readObjectAtOffset(off_t offset, int exp_objid, int exp_generation,
     }
     catch (QPDFExc& e)
     {
-	if (exp_objid && this->attempt_recovery)
+	if (exp_objid && try_recovery && this->attempt_recovery)
 	{
 	    // Try again after reconstructing xref table
 	    reconstruct_xref(e);
@@ -1414,13 +1532,27 @@ QPDF::readObjectAtOffset(off_t offset, int exp_objid, int exp_generation,
 		(this->xref_table[og].getType() == 1))
 	    {
 		off_t new_offset = this->xref_table[og].getOffset();
-		// Call readObjectAtOffset with 0 for exp_objid to
-		// avoid an infinite loop.
-		QPDFObjectHandle result =
-		    readObjectAtOffset(new_offset, 0, 0, objid, generation);
+		QPDFObjectHandle result = readObjectAtOffset(
+		    false, new_offset, description,
+		    exp_objid, exp_generation, objid, generation);
 		QTC::TC("qpdf", "QPDF recovered in readObjectAtOffset");
 		return result;
 	    }
+	    else
+	    {
+		QTC::TC("qpdf", "QPDF object gone after xref reconstruction");
+		warn(QPDFExc(
+			 qpdf_e_damaged_pdf, this->file.getName(),
+			 "", 0,
+			 std::string(
+			     "object " +
+			     QUtil::int_to_string(exp_objid) +
+			     " " +
+			     QUtil::int_to_string(exp_generation) +
+			     " not found in file after regenerating"
+			     " cross reference table")));
+		return QPDFObjectHandle::newNull();
+	    }
 	}
 	else
 	{
@@ -1429,13 +1561,14 @@ QPDF::readObjectAtOffset(off_t offset, int exp_objid, int exp_generation,
     }
 
     QPDFObjectHandle oh = readObject(
-	&this->file, objid, generation, false);
+	&this->file, description, objid, generation, false);
 
     if (! (readToken(&this->file) ==
 	   QPDFTokenizer::Token(QPDFTokenizer::tt_word, "endobj")))
     {
 	QTC::TC("qpdf", "QPDF err expected endobj");
-	warn(QPDFExc(this->file.getName(), this->file.getLastOffset(),
+	warn(QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+		     this->last_object_description, this->file.getLastOffset(),
 		     "expected endobj"));
     }
 
@@ -1460,7 +1593,7 @@ QPDF::readObjectAtOffset(off_t offset, int exp_objid, int exp_generation,
 	    char ch;
 	    if (this->file.read(&ch, 1))
 	    {
-		if (! isspace(ch))
+		if (! isspace((unsigned char)ch))
 		{
 		    this->file.seek(-1, SEEK_CUR);
 		    break;
@@ -1468,7 +1601,8 @@ QPDF::readObjectAtOffset(off_t offset, int exp_objid, int exp_generation,
 	    }
 	    else
 	    {
-		throw QPDFExc(this->file.getName(), offset,
+		throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+			      this->last_object_description, offset,
 			      "EOF after endobj");
 	    }
 	}
@@ -1507,7 +1641,7 @@ QPDF::resolve(int objid, int generation)
 		int aobjid;
 		int ageneration;
 		QPDFObjectHandle oh =
-		    readObjectAtOffset(offset, objid, generation,
+		    readObjectAtOffset(true, offset, "", objid, generation,
 				       aobjid, ageneration);
 	    }
 	    break;
@@ -1517,7 +1651,7 @@ QPDF::resolve(int objid, int generation)
 	    break;
 
 	  default:
-	    throw QPDFExc(this->file.getName(), 0,
+	    throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(), "", 0,
 			  "object " +
 			  QUtil::int_to_string(objid) + "/" +
 			  QUtil::int_to_string(generation) +
@@ -1535,7 +1669,9 @@ QPDF::resolveObjectsInStream(int obj_stream_number)
     QPDFObjectHandle obj_stream = getObjectByID(obj_stream_number, 0);
     if (! obj_stream.isStream())
     {
-	throw QPDFExc(this->file.getName(), this->file.getLastOffset(),
+	throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+		      this->last_object_description,
+		      this->file.getLastOffset(),
 		      "supposed object stream " +
 		      QUtil::int_to_string(obj_stream_number) +
 		      " is not a stream");
@@ -1551,7 +1687,10 @@ QPDF::resolveObjectsInStream(int obj_stream_number)
     if (! (dict.getKey("/Type").isName() &&
 	   dict.getKey("/Type").getName() == "/ObjStm"))
     {
-	throw QPDFExc(this->file.getName(), this->file.getLastOffset(),
+	QTC::TC("qpdf", "QPDF ERR object stream with wrong type");
+	throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+		      this->last_object_description,
+		      this->file.getLastOffset(),
 		      "supposed object stream " +
 		      QUtil::int_to_string(obj_stream_number) +
 		      " has wrong type");
@@ -1560,7 +1699,9 @@ QPDF::resolveObjectsInStream(int obj_stream_number)
     if (! (dict.getKey("/N").isInteger() &&
 	   dict.getKey("/First").isInteger()))
     {
-	throw QPDFExc(this->file.getName(), this->file.getLastOffset(),
+	throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+		      this->last_object_description,
+		      this->file.getLastOffset(),
 		      "object stream " +
 		      QUtil::int_to_string(obj_stream_number) +
 		      " has incorrect keys");
@@ -1583,7 +1724,8 @@ QPDF::resolveObjectsInStream(int obj_stream_number)
 	if (! ((tnum.getType() == QPDFTokenizer::tt_integer) &&
 	       (toffset.getType() == QPDFTokenizer::tt_integer)))
 	{
-	    throw QPDFExc(input.getName(), input.getLastOffset(),
+	    throw QPDFExc(qpdf_e_damaged_pdf, input.getName(),
+			  this->last_object_description, input.getLastOffset(),
 			  "expected integer in object stream header");
 	}
 
@@ -1598,7 +1740,7 @@ QPDF::resolveObjectsInStream(int obj_stream_number)
 	int obj = (*iter).first;
 	int offset = (*iter).second;
 	input.seek(offset, SEEK_SET);
-	QPDFObjectHandle oh = readObject(&input, obj, 0, true);
+	QPDFObjectHandle oh = readObject(&input, "", obj, 0, true);
 
 	// Store in cache
 	ObjGen og(obj, 0);
@@ -1798,17 +1940,7 @@ QPDF::pipeStreamData(int objid, int generation,
     std::vector<PointerHolder<Pipeline> > to_delete;
     if (this->encrypted)
     {
-	bool xref_stream = false;
-	if (stream_dict.getKey("/Type").isName() &&
-	    (stream_dict.getKey("/Type").getName() == "/XRef"))
-	{
-	    QTC::TC("qpdf", "QPDF piping xref stream from encrypted file");
-	    xref_stream = true;
-	}
-	if (! xref_stream)
-	{
-	    decryptStream(pipeline, objid, generation, to_delete);
-	}
+	decryptStream(pipeline, objid, generation, stream_dict, to_delete);
     }
 
     try
@@ -1821,20 +1953,28 @@ QPDF::pipeStreamData(int objid, int generation,
 	    size_t len = this->file.read(buf, to_read);
 	    if (len == 0)
 	    {
-		throw QPDFExc(this->file.getName(), this->file.getLastOffset(),
+		throw QPDFExc(qpdf_e_damaged_pdf,
+			      this->file.getName(),
+			      this->last_object_description,
+			      this->file.getLastOffset(),
 			      "unexpected EOF reading stream data");
 	    }
 	    length -= len;
 	    pipeline->write((unsigned char*)buf, len);
 	}
     }
-    catch (QEXC::General& e)
+    catch (QPDFExc& e)
+    {
+	warn(e);
+    }
+    catch (std::runtime_error& e)
     {
 	QTC::TC("qpdf", "QPDF decoding error warning");
-	warn(QPDFExc(this->file.getName(), this->file.getLastOffset(),
+	warn(QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+		     "", this->file.getLastOffset(),
 		     "error decoding stream data for object " +
 		     QUtil::int_to_string(objid) + " " +
-		     QUtil::int_to_string(generation) + ": " + e.unparse()));
+		     QUtil::int_to_string(generation) + ": " + e.what()));
     }
     pipeline->finish();
 }
@@ -1887,6 +2027,9 @@ QPDF::getAllPagesInternal(QPDFObjectHandle cur_pages,
     }
     else
     {
-	throw QPDFExc(this->file.getName() + ": invalid Type in page tree");
+	throw QPDFExc(qpdf_e_damaged_pdf, this->file.getName(),
+		      this->last_object_description,
+		      this->file.getLastOffset(),
+		      + ": invalid Type in page tree");
     }
 }
